@@ -10,8 +10,8 @@ from typing import Any
 
 from .iam_common import (
     IamServer, IamParam, UserParam, _iam_lock,
-    _get_iam_users, _get_iam_registry,  # _get_public_key,
-    _get_login_timeout, _get_user_data
+    _get_iam_users, _get_iam_registry, _get_public_key,
+    _get_login_timeout, _get_user_data, _iam_server_from_issuer
 )
 from .token_pomes import token_get_claims, token_validate
 
@@ -368,7 +368,7 @@ def __assert_link(iam_server: IamServer,
     """
     Make sure *iam_server* has a link associating *user_id* to an internal user identification.
     This is a requirement for exchanging a token issued by a federated *IAM* server for an equivalent
-    one from *iam_server.
+    one from *iam_server*.
 
     :param iam_server: the reference *IAM* server
     :param user_id: the reference user identification
@@ -376,6 +376,9 @@ def __assert_link(iam_server: IamServer,
     :param errors: incidental errors
     :param logger: optional logger
     """
+    if logger:
+        logger.debug(msg="Verifying associations for user "
+                         f"'{user_id}' in IAM server '{iam_server}'")
     # obtain a token with administrative rights
     admin_token: str = __get_administrative_token(iam_server=iam_server,
                                                   errors=errors,
@@ -384,7 +387,10 @@ def __assert_link(iam_server: IamServer,
         registry: dict[str, Any] = _get_iam_registry(iam_server=iam_server,
                                                      errors=errors,
                                                      logger=logger)
-        # obtain the internal user identification for 'from_id'
+        # obtain the internal user identification for 'user_id'
+        if logger:
+            logger.debug(msg="Obtaining internal identification "
+                             f"for user {user_id} in IAM server {iam_server}")
         url: str = f"{registry[IamParam.URL_BASE]}/admin/realms/{registry[IamParam.CLIENT_REALM]}/users"
         header_data: dict[str, str] = {
             "Authorization": f"Bearer {admin_token}",
@@ -394,15 +400,18 @@ def __assert_link(iam_server: IamServer,
             "username": user_id,
             "exact": "true"
         }
-        users: dict[str, Any] = __get_for_data(url=url,
-                                               header_data=header_data,
-                                               params=params,
-                                               errors=errors,
-                                               logger=logger)
+        users: list[dict[str, Any]] = __get_for_data(url=url,
+                                                     header_data=header_data,
+                                                     params=params,
+                                                     errors=errors,
+                                                     logger=logger)
         if users:
             # verify whether the 'oidc' protocol is referred to in an
-            # association between 'from_id' and the internal user identification
-            internal_id: str = users.get("id")
+            # association between 'user_id' and the internal user identification
+            internal_id: str = users[0].get("id")
+            if logger:
+                logger.debug(msg="Obtaining the providers federated with "
+                                 f"IAM server '{iam_server}' for internal identification '{internal_id}'")
             url = (f"{registry[IamParam.URL_BASE]}/admin/realms/"
                    f"{registry[IamParam.CLIENT_REALM]}/users/{internal_id}/federated-identity")
             providers: list[dict[str, Any]] = __get_for_data(url=url,
@@ -411,27 +420,38 @@ def __assert_link(iam_server: IamServer,
                                                              errors=errors,
                                                              logger=logger)
             no_link: bool = True
-            for provider in providers:
-                if provider.get("identityProvider") == "oidc":
-                    no_link = False
-                    break
-            if no_link:
-                # link the identities
-                claims: dict[str, dict[str, Any]] = token_get_claims(token=token,
-                                                                     errors=errors,
-                                                                     logger=logger)
-                if claims:
-                    token_sub: str = claims["paylod"]["sub"]
-                    url += "/oidc"
-                    body_data: dict[str, Any] = {
-                        "userId": token_sub,
-                        "userName": user_id
-                    }
-                    __post_data(url=url,
-                                header_data=header_data,
-                                body_data=body_data,
-                                errors=errors,
-                                logger=logger)
+            claims: dict[str, dict[str, Any]] = token_get_claims(token=token,
+                                                                 errors=errors,
+                                                                 logger=logger)
+            issuer: str = claims["payload"]["iss"] if claims else None
+            provider_name: str = _iam_server_from_issuer(issuer=issuer,
+                                                         errors=errors,
+                                                         logger=logger) if issuer else None
+            if provider_name:
+                for provider in providers:
+                    if provider.get("identityProvider") == provider_name:
+                        no_link = False
+                        break
+                if no_link:
+                    # link the identities
+                    claims: dict[str, dict[str, Any]] = token_get_claims(token=token,
+                                                                         errors=errors,
+                                                                         logger=logger)
+                    if claims:
+                        token_sub: str = claims["payload"]["sub"]
+                        if logger:
+                            logger.debug(msg="Creating an association between identifications "
+                                             f"'{user_id}' and '{token_sub}' in IAM server {iam_server}")
+                        url += f"/{provider_name}"
+                        body_data: dict[str, Any] = {
+                            "userId": token_sub,
+                            "userName": user_id
+                        }
+                        __post_data(url=url,
+                                    header_data=header_data,
+                                    body_data=body_data,
+                                    errors=errors,
+                                    logger=logger)
 
 
 def __get_administrative_token(iam_server: IamServer,
@@ -451,6 +471,10 @@ def __get_administrative_token(iam_server: IamServer,
     # initialize the return variable
     result: str | None = None
 
+    if logger:
+        logger.debug(msg="Requesting a token with "
+                         f"administrative rights to IAM Server '{iam_server}'")
+
     # obtain the IAM server's registry
     registry: dict[str, Any] = _get_iam_registry(iam_server=iam_server,
                                                  errors=errors,
@@ -461,10 +485,10 @@ def __get_administrative_token(iam_server: IamServer,
                 "Content-Type": "application/x-www-form-urlencoded"
             }
             body_data: dict[str, str] = {
-                "grant-type": "password",
+                "grant_type": "password",
                 "username": registry[IamParam.ADMIN_ID],
                 "password": registry[IamParam.ADMIN_SECRET],
-                "client_id": registry[IamParam.CLIENT_ID]
+                "client_id": "admin-cli"
             }
             token_data: dict[str, Any] = __post_for_token(iam_server=iam_server,
                                                           header_data=header_data,
@@ -474,10 +498,13 @@ def __get_administrative_token(iam_server: IamServer,
             if token_data:
                 # obtain the token
                 result = token_data["access_token"]
+                if logger:
+                    logger.debug(msg="Administrative token obtained")
 
         elif logger or isinstance(errors, list):
             msg: str = ("Credentials for administrator of realm "
-                        f"'{registry[IamParam.CLIENT_REALM]}' of '{iam_server}' not provided")
+                        f"'{registry[IamParam.CLIENT_REALM]}' "
+                        f"at IAM server '{iam_server}' not provided")
             if logger:
                 logger.error(msg=msg)
             if isinstance(errors, list):
@@ -520,14 +547,19 @@ def __get_client_secret(iam_server: IamServer,
                                                 errors=errors,
                                                 logger=logger)
         if token:
+            realm: str = registry[IamParam.CLIENT_REALM]
+            client_id: str = registry[IamParam.CLIENT_ID]
+            if logger:
+                logger.debug(msg=f"Obtaining the UUID for client '{client_id}', "
+                                 f"in realm '{realm}' at IAM server '{iam_server}'")
             # obtain the client UUID
-            url: str = f"{registry[IamParam.URL_BASE]}/realms/{registry[IamParam.CLIENT_REALM]}/clients"
+            url: str = f"{registry[IamParam.URL_BASE]}/realms/{realm}/clients"
             header_data: dict[str, str] = {
                 "Authorization": f"Bearer {token}",
                 "Content-Type": "application/json"
             }
             params: dict[str, str] = {
-                "clientId": registry[IamParam.CLIENT_ID]
+                "clientId": client_id
             }
             clients: list[dict[str, Any]] = __get_for_data(url=url,
                                                            header_data=header_data,
@@ -537,6 +569,9 @@ def __get_client_secret(iam_server: IamServer,
             if clients:
                 # obtain the client's secret password
                 client_uuid: str = clients[0]["id"]
+                if logger:
+                    logger.debug(msg=f"Obtaining the secret for client UUID '{client_uuid}', "
+                                     f"in realm '{realm}' at IAM server '{iam_server}'")
                 url += f"/{client_uuid}/client-secret"
                 reply: dict[str, Any] = __get_for_data(url=url,
                                                        header_data=header_data,
@@ -675,9 +710,9 @@ def __post_for_token(iam_server: IamServer,
         - "username": <realm-administrator-identification>
         - "password": <realm-administrator-secret>
 
-    These attributes are then added to *body_data*:
+    These attributes are then added to *body_data*, except for acquiring administrative tokens:
         - "client_id": <client-id>
-        - "client_secret": <client-secret>  <- except for acquiring administrative tokens
+        - "client_secret": <client-secret>
 
     If the operation is successful, the token data is stored in the *IAM* server's registry, and returned.
     Otherwise, *errors* will contain the appropriate error message.
@@ -709,21 +744,25 @@ def __post_for_token(iam_server: IamServer,
                                                      logger=logger)
         if registry:
             # complete the data to send in body of request
-            body_data["client_id"] = registry[IamParam.CLIENT_ID]
+            if body_data["grant_type"] != "password":
+                body_data["client_id"] = registry[IamParam.CLIENT_ID]
 
             # build the URL
             base_url: str = f"{registry[IamParam.URL_BASE]}/realms/{registry[IamParam.CLIENT_REALM]}"
             url: str = f"{base_url}/protocol/openid-connect/token"
-
-            # log the POST ('client_secret' data must not be shown in log)
+            #  'client_secret' data must not be shown in log
+            msg: str = f"POST {url}, {json.dumps(obj=body_data,
+                                                 ensure_ascii=False)}"
+            if body_data["grant_type"] != "password":
+                # 'client_secret' not required for requesting tokens from staging environments
+                client_secret: str = __get_client_secret(iam_server=iam_server,
+                                                         errors=None,
+                                                         logger=logger)
+                if client_secret:
+                    body_data["client_secret"] = client_secret
+            # log the POST
             if logger:
-                logger.debug(msg=f"POST {url}, {json.dumps(obj=body_data,
-                                                           ensure_ascii=False)}")
-            client_secret: str = __get_client_secret(iam_server=iam_server,
-                                                     errors=errors,
-                                                     logger=logger)
-            if body_data["grant_type"] != "password" and client_secret:
-                body_data["client_secret"] = client_secret
+                logger.debug(msg=msg)
 
             # obtain the token
             try:
@@ -806,9 +845,9 @@ def __validate_and_store(iam_server: IamServer,
             user_data["access-expiration"] = now + token_data.get("expires_in")
             refresh_exp: int = user_data.get("refresh_expires_in")
             user_data["refresh-expiration"] = (now + refresh_exp) if refresh_exp else sys.maxsize
-            # public_key: str = _get_public_key(iam_server=iam_server,
-            #                                   errors=errors,
-            #                                   logger=logger)
+            public_key: str = _get_public_key(iam_server=iam_server,
+                                              errors=errors,
+                                              logger=logger)
             recipient_attr = registry[IamParam.RECIPIENT_ATTR]
             login_id = user_data.pop("login-id", None)
             base_url: str = f"{registry[IamParam.URL_BASE]}/realms/{registry[IamParam.CLIENT_REALM]}"
@@ -816,7 +855,7 @@ def __validate_and_store(iam_server: IamServer,
                                                                issuer=base_url,
                                                                recipient_id=login_id,
                                                                recipient_attr=recipient_attr,
-                                                               # public_key=public_key,
+                                                               public_key=public_key,
                                                                errors=errors,
                                                                logger=logger)
             if claims:
