@@ -40,7 +40,7 @@ def iam_setup(flask_app: Flask,
     the first time it is needed.
 
     :param flask_app: the Flask application
-    :param iam_server: identifies the supported *IAM* server (*jusbr* or *keycloak*)
+    :param iam_server: identifies the supported *IAM* server (currently, *jusbr* or *keycloak*)
     :param base_url: base URL to request services
     :param client_id: the client's identification with the *IAM* server
     :param client_realm: the client realm
